import git
import json
import os
import shutil
import sys
import tarfile
import time
import tempfile
from datetime import datetime
from siliconcompiler.utils import get_file_template
from siliconcompiler.scheduler import _makecmd, _write_task_manifest, _get_machine_info
from siliconcompiler.tools._common import get_tool_task


def generate_testcase(chip,
                      step,
                      index,
                      archive_name=None,
                      archive_directory=None,
                      include_pdks=True,
                      include_specific_pdks=None,
                      include_libraries=True,
                      include_specific_libraries=None,
                      hash_files=False,
                      verbose_collect=True):

    # Save original schema since it will be modified
    schema_copy = chip.schema.copy()

    issue_dir = tempfile.TemporaryDirectory(prefix='sc_issue_')

    chip.set('option', 'continue', True)
    if hash_files:
        for key in chip.allkeys():
            if key[0] == 'history':
                continue
            if len(key) > 1:
                if key[-2] == 'option' and key[-1] == 'builddir':
                    continue
                if key[-2] == 'option' and key[-1] == 'cachedir':
                    continue
            sc_type = chip.get(*key, field='type')
            if 'file' not in sc_type and 'dir' not in sc_type:
                continue
            for _, key_step, key_index in chip.schema._getvals(*key):
                chip.hash_files(*key,
                                check=False,
                                allow_cache=True,
                                verbose=False,
                                skip_missing=True,
                                step=key_step, index=key_index)

    manifest_path = os.path.join(issue_dir.name, 'orig_manifest.json')
    chip.write_manifest(manifest_path)

    flow = chip.get('option', 'flow')
    tool, task = get_tool_task(chip, step, index, flow=flow)
    task_requires = chip.get('tool', tool, 'task', task, 'require',
                             step=step, index=index)

    def determine_copy(*keypath, in_require):
        copy = in_require

        if keypath[0] == 'library':
            # only copy libraries if selected
            if include_specific_libraries and keypath[1] in include_specific_libraries:
                copy = True
            else:
                copy = include_libraries

            copy = copy and determine_copy(*keypath[2:], in_require=in_require)
        elif keypath[0] == 'pdk':
            # only copy pdks if selected
            if include_specific_pdks and keypath[1] in include_specific_pdks:
                copy = True
            else:
                copy = include_pdks
        elif keypath[0] == 'history':
            # Skip history
            copy = False
        elif keypath[0] == 'package':
            # Skip packages
            copy = False
        elif keypath[0] == 'tool':
            # Only grab tool / tasks
            copy = False
            if list(keypath[0:4]) == ['tool', tool, 'task', task]:
                # Get files associated with testcase tool / task
                copy = True
                if len(keypath) >= 5:
                    if keypath[4] in ('output', 'input', 'report'):
                        # Skip input, output, and report files
                        copy = False
        elif keypath[0] == 'option':
            if keypath[1] == 'builddir':
                # Avoid build directory
                copy = False
            elif keypath[1] == 'cachedir':
                # Avoid cache directory
                copy = False
            elif keypath[1] == 'cfg':
                # Avoid all of cfg, since we are getting the manifest separately
                copy = False
            elif keypath[1] == 'credentials':
                # Exclude credentials file
                copy = False

        return copy

    for keypath in chip.allkeys():
        if 'default' in keypath:
            continue

        sctype = chip.get(*keypath, field='type')
        if 'file' not in sctype and 'dir' not in sctype:
            continue

        chip.set(*keypath,
                 determine_copy(*keypath,
                                in_require=','.join(keypath) in task_requires),
                 field='copy')

    # Collect files
    work_dir = chip.getworkdir(step=step, index=index)

    builddir = chip.get('option', 'builddir')
    if os.path.isabs(builddir):
        # If build is an abs path, grab last directory
        chip.set('option', 'builddir', os.path.basename(builddir))

    # Temporarily change current directory to appear to be issue_dir
    original_cwd = chip.cwd
    chip.cwd = issue_dir.name

    # Get new directories
    job_dir = chip.getworkdir()
    new_work_dir = chip.getworkdir(step=step, index=index)
    collection_dir = chip._getcollectdir()

    # Restore current directory
    chip.cwd = original_cwd

    # Copy in issue run files
    shutil.copytree(work_dir, new_work_dir, dirs_exist_ok=True)
    # Copy in source files
    chip.collect(directory=collection_dir, verbose=verbose_collect)

    # Set relative path to generate runnable files
    chip._relative_path = new_work_dir
    chip.cwd = issue_dir.name

    current_work_dir = os.getcwd()
    os.chdir(new_work_dir)

    # Rewrite replay.sh
    prev_quiet = chip.get('option', 'quiet', step=step, index=index)
    chip.set('option', 'quiet', True, step=step, index=index)
    from siliconcompiler import SiliconCompilerError
    try:
        # Rerun setup
        chip.set('arg', 'step', step)
        chip.set('arg', 'index', index)
        func = getattr(chip._get_task_module(step, index, flow=flow), 'pre_process', None)
        if func:
            try:
                # Rerun pre_process
                func(chip)
            except Exception:
                pass
    except SiliconCompilerError:
        pass
    chip.set('option', 'quiet', prev_quiet, step=step, index=index)

    flow = chip.get('option', 'flow')
    is_python_tool = hasattr(chip._get_task_module(step, index, flow=flow), 'run')

    if not is_python_tool:
        _makecmd(chip,
                 tool, task, step, index,
                 script_name=f'{chip.getworkdir(step=step, index=index)}/replay.sh',
                 include_path=False)

    # Rewrite tool manifest
    chip.set('arg', 'step', step)
    chip.set('arg', 'index', index)
    _write_task_manifest(chip, tool, path=new_work_dir)

    # Restore normal path behavior
    chip._relative_path = None

    # Restore current directory
    chip.cwd = original_cwd
    os.chdir(current_work_dir)

    git_data = {}
    try:
        # Check git information
        repo = git.Repo(path=os.path.join(chip.scroot, '..'))
        commit = repo.head.commit
        git_data['commit'] = commit.hexsha
        git_data['date'] = time.strftime('%Y-%m-%d %H:%M:%S',
                                         time.gmtime(commit.committed_date))
        git_data['author'] = f'{commit.author.name} <{commit.author.email}>'
        git_data['msg'] = commit.message
        # Count number of commits ahead of version
        version_tag = repo.tag(f'v{chip.scversion}')
        count = 0
        for c in commit.iter_parents():
            count += 1
            if c == version_tag.commit:
                break
        git_data['count'] = count
    except git.InvalidGitRepositoryError:
        pass
    except Exception as e:
        git_data['failed'] = str(e)
        pass

    tool, task = get_tool_task(chip, step=step, index=index)

    issue_time = time.time()
    issue_information = {}
    issue_information['environment'] = {key: value for key, value in os.environ.items()}
    issue_information['python'] = {"path": sys.path,
                                   "version": sys.version}
    issue_information['date'] = datetime.fromtimestamp(issue_time).strftime('%Y-%m-%d %H:%M:%S')
    issue_information['machine'] = _get_machine_info()
    issue_information['run'] = {'step': step,
                                'index': index,
                                'libraries_included': include_libraries,
                                'pdks_included': include_pdks,
                                'tool': tool,
                                'toolversion': chip.get('record', 'toolversion',
                                                        step=step, index=index),
                                'task': task}
    issue_information['version'] = {'schema': chip.schemaversion,
                                    'sc': chip.scversion,
                                    'git': git_data}

    if not archive_name:
        design = chip.design
        job = chip.get('option', 'jobname')
        file_time = datetime.fromtimestamp(issue_time).strftime('%Y%m%d-%H%M%S')
        archive_name = f'sc_issue_{design}_{job}_{step}{index}_{file_time}.tar.gz'

    # Make support files
    issue_path = os.path.join(issue_dir.name, 'issue.json')
    with open(issue_path, 'w') as fd:
        json.dump(issue_information, fd, indent=4, sort_keys=True)

    readme_path = os.path.join(issue_dir.name, 'README.txt')
    with open(readme_path, 'w') as f:
        f.write(get_file_template('issue/README.txt').render(
            archive_name=archive_name,
            has_run=not is_python_tool,
            **issue_information))

    run_path = None
    if not is_python_tool:
        run_path = os.path.join(issue_dir.name, 'run.sh')
        with open(run_path, 'w') as f:
            replay_dir = os.path.relpath(chip.getworkdir(step=step, index=index),
                                         chip.cwd)
            issue_title = f'{chip.design} for {step}{index} using {tool}/{task}'
            f.write(get_file_template('issue/run.sh').render(
                title=issue_title,
                exec_dir=replay_dir
            ))
        os.chmod(run_path, 0o755)

    full_archive_path = archive_name
    if archive_directory:
        full_archive_path = os.path.join(archive_directory, archive_name)
    full_archive_path = os.path.abspath(full_archive_path)
    # Build archive
    arch_base_dir = os.path.basename(archive_name).split('.')[0]
    with tarfile.open(full_archive_path, "w:gz") as tar:
        # Add individual files
        add_files = [manifest_path,
                     issue_path,
                     readme_path]
        if not is_python_tool:
            add_files.append(run_path)
        for path in add_files:
            tar.add(os.path.abspath(path),
                    arcname=os.path.join(arch_base_dir,
                                         os.path.basename(path)))

        tar.add(job_dir,
                arcname=os.path.join(arch_base_dir,
                                     os.path.relpath(job_dir,
                                                     issue_dir.name)))

    issue_dir.cleanup()

    chip.logger.info(f'Generated testcase for {step}{index} in: '
                     f'{full_archive_path}')

    # Restore original schema
    chip.schema = schema_copy



import os
import math
from siliconcompiler import SiliconCompilerError, NodeStatus
from siliconcompiler.tools._common import input_file_node_name, get_tool_task


def _nodes_to_execute(chip, flow, from_nodes, to_nodes, prune_nodes):
    '''
    Assumes a flowgraph with valid edges for the inputs
    '''
    nodes_to_execute = []
    for from_node in from_nodes:
        for node in _nodes_to_execute_recursive(chip, flow, from_node, to_nodes, prune_nodes):
            if node not in nodes_to_execute:
                nodes_to_execute.append(node)
    return nodes_to_execute


def _nodes_to_execute_recursive(chip, flow, from_node, to_nodes, prune_nodes, path=[]):
    path = path.copy()
    nodes_to_execute = []

    if from_node in prune_nodes:
        return []
    if from_node in path:
        raise SiliconCompilerError(f'Path {path} would form a circle with {from_node}')
    path.append(from_node)

    if from_node in to_nodes:
        for node in path:
            nodes_to_execute.append(node)
    for output_node in _get_flowgraph_node_outputs(chip, flow, from_node):
        for node in _nodes_to_execute_recursive(chip, flow, output_node, to_nodes,
                                                prune_nodes, path=path):
            if node not in nodes_to_execute:
                nodes_to_execute.append(node)

    return nodes_to_execute


def _unreachable_steps_to_execute(chip, flow, cond=lambda _: True):
    from_nodes = set(_get_execution_entry_nodes(chip, flow))
    to_nodes = set(_get_execution_exit_nodes(chip, flow))
    prune_nodes = chip.get('option', 'prune')
    reachable_nodes = set(_reachable_flowgraph_nodes(chip, flow, from_nodes, cond=cond,
                                                     prune_nodes=prune_nodes))
    unreachable_nodes = to_nodes.difference(reachable_nodes)
    unreachable_steps = set()
    for unreachable_node in unreachable_nodes:
        if not any(filter(lambda node: node[0] == unreachable_node[0], reachable_nodes)):
            unreachable_steps.add(unreachable_node[0])
    return unreachable_steps


def _reachable_flowgraph_nodes(chip, flow, from_nodes, cond=lambda _: True, prune_nodes=[]):
    visited_nodes = set()
    current_nodes = from_nodes.copy()
    while current_nodes:
        current_nodes_copy = current_nodes.copy()
        for current_node in current_nodes_copy:
            if current_node in prune_nodes:
                current_nodes.remove(current_node)
                continue
            if cond(current_node):
                visited_nodes.add(current_node)
                current_nodes.remove(current_node)
                outputs = _get_flowgraph_node_outputs(chip, flow, current_node)
                current_nodes.update(outputs)
        if current_nodes == current_nodes_copy:
            break
    return visited_nodes


def _get_flowgraph_node_inputs(chip, flow, node):
    step, index = node
    inputs = set()
    for in_node in chip.get('flowgraph', flow, step, index, 'input'):
        if chip.get('record', 'status', step=in_node[0], index=in_node[1]) == \
                NodeStatus.SKIPPED:
            inputs.update(_get_flowgraph_node_inputs(chip, flow, in_node))
        else:
            inputs.add(in_node)
    return list(inputs)


def _get_pruned_flowgraph_nodes(chip, flow, prune_nodes):
    # Ignore option from/to, we want reachable nodes of the whole flowgraph
    from_nodes = set(_get_flowgraph_entry_nodes(chip, flow))
    return _reachable_flowgraph_nodes(chip, flow, from_nodes, prune_nodes=prune_nodes)


def _get_pruned_node_inputs(chip, flow, node):
    prune_nodes = chip.get('option', 'prune')
    pruned_flowgraph_nodes = _get_pruned_flowgraph_nodes(chip, flow, prune_nodes)
    return list(filter(lambda node: node in pruned_flowgraph_nodes,
                       _get_flowgraph_node_inputs(chip, flow, node)))


def _get_flowgraph_node_outputs(chip, flow, node):
    node_outputs = []

    iter_nodes = _get_flowgraph_nodes(chip, flow)
    for iter_node in iter_nodes:
        iter_node_inputs = chip.get('flowgraph', flow, *iter_node, 'input')
        if node in iter_node_inputs:
            node_outputs.append(iter_node)

    return node_outputs


def _get_flowgraph_nodes(chip, flow, steps=None, indices=None):
    nodes = []
    for step in chip.getkeys('flowgraph', flow):
        if steps and step not in steps:
            continue
        for index in chip.getkeys('flowgraph', flow, step):
            if indices and index not in indices:
                continue
            nodes.append((step, index))
    return nodes


#######################################
def _get_execution_entry_nodes(chip, flow):
    if chip.get('arg', 'step') and chip.get('arg', 'index'):
        return [(chip.get('arg', 'step'), chip.get('arg', 'index'))]
    if chip.get('arg', 'step'):
        return _get_flowgraph_nodes(chip, flow, steps=[chip.get('arg', 'step')])
    # If we explicitly get the nodes for a flow other than the current one,
    # Ignore the 'option' 'from'
    if chip.get('option', 'flow') == flow and chip.get('option', 'from'):
        return _get_flowgraph_nodes(chip, flow, steps=chip.get('option', 'from'))
    return _get_flowgraph_entry_nodes(chip, flow)


def _get_flowgraph_entry_nodes(chip, flow, steps=None):
    '''
    Collect all step/indices that represent the entry
    nodes for the flowgraph
    '''
    nodes = []
    for (step, index) in _get_flowgraph_nodes(chip, flow, steps=steps):
        if not chip.get('flowgraph', flow, step, index, 'input'):
            nodes.append((step, index))
    return nodes


def _get_execution_exit_nodes(chip, flow):
    if chip.get('arg', 'step') and chip.get('arg', 'index'):
        return [(chip.get('arg', 'step'), chip.get('arg', 'index'))]
    if chip.get('arg', 'step'):
        return _get_flowgraph_nodes(chip, flow, steps=[chip.get('arg', 'step')])
    # If we explicitly get the nodes for a flow other than the current one,
    # Ignore the 'option' 'to'
    if chip.get('option', 'flow') == flow and chip.get('option', 'to'):
        return _get_flowgraph_nodes(chip, flow, steps=chip.get('option', 'to'))
    return _get_flowgraph_exit_nodes(chip, flow)


#######################################
def _get_flowgraph_exit_nodes(chip, flow, steps=None):
    '''
    Collect all step/indices that represent the exit
    nodes for the flowgraph
    '''
    inputnodes = []
    for (step, index) in _get_flowgraph_nodes(chip, flow, steps=steps):
        inputnodes.extend(chip.get('flowgraph', flow, step, index, 'input'))
    nodes = []
    for (step, index) in _get_flowgraph_nodes(chip, flow, steps=steps):
        if (step, index) not in inputnodes:
            nodes.append((step, index))
    return nodes


#######################################
def _get_flowgraph_execution_order(chip, flow, reverse=False):
    '''
    Generates a list of nodes in the order they will be executed.
    '''

    # Generate execution edges lookup map
    ex_map = {}
    for step, index in _get_flowgraph_nodes(chip, flow):
        for istep, iindex in chip.get('flowgraph', flow, step, index, 'input'):
            if reverse:
                ex_map.setdefault((step, index), set()).add((istep, iindex))
            else:
                ex_map.setdefault((istep, iindex), set()).add((step, index))

    rev_ex_map = {}
    for node, edges in ex_map.items():
        for step, index in edges:
            rev_ex_map.setdefault((step, index), set()).add(node)

    # Collect execution order of nodes
    if reverse:
        order = [set(_get_flowgraph_exit_nodes(chip, flow))]
    else:
        order = [set(_get_flowgraph_entry_nodes(chip, flow))]

    visited = set()
    while True:
        next_level = set()
        next_visited = set()
        for step, index in sorted(order[-1]):
            if (step, index) not in rev_ex_map:
                # No edges so assume inputs are okay
                inputs_valid = True
            else:
                inputs_valid = all([node in visited for node in rev_ex_map[(step, index)]])

            if inputs_valid:
                next_visited.add((step, index))
                if (step, index) in ex_map:
                    next_level.update(ex_map.pop((step, index)))
            else:
                next_level.add((step, index))

        visited.update(next_visited)

        if not next_level:
            break

        order.append(next_level)

    # Filter duplicates from flow
    used_nodes = set()
    exec_order = []
    order.reverse()
    for n, level_nodes in enumerate(order):
        exec_order.append(list(level_nodes.difference(used_nodes)))
        used_nodes.update(level_nodes)

    exec_order.reverse()

    return [sorted(level) for level in exec_order]


def get_executed_nodes(chip, flow):
    from_nodes = _get_flowgraph_entry_nodes(chip, flow)
    return get_nodes_from(chip, flow, from_nodes)


def get_nodes_from(chip, flow, nodes):
    to_nodes = _get_execution_exit_nodes(chip, flow)
    return _nodes_to_execute(chip,
                             flow,
                             set(nodes),
                             set(to_nodes),
                             set(chip.get('option', 'prune')))


###########################################################################
def nodes_to_execute(chip, flow=None):
    '''
    Returns an ordered list of flowgraph nodes which will be executed.
    This takes the from/to options into account if flow is the current flow or None.

    Returns:
        A list of nodes that will get executed during run() (or a specific flow).

    Example:
        >>> nodes = nodes_to_execute()
    '''
    if flow is None:
        flow = chip.get('option', 'flow')

    from_nodes = set(_get_execution_entry_nodes(chip, flow))
    to_nodes = set(_get_execution_exit_nodes(chip, flow))
    prune_nodes = set(chip.get('option', 'prune'))
    if from_nodes == to_nodes:
        return list(filter(lambda node: node not in prune_nodes, from_nodes))
    return _nodes_to_execute(chip, flow, from_nodes, to_nodes, prune_nodes)


###########################################################################
def _check_flowgraph(chip, flow=None):
    '''
    Check if flowgraph is valid.

    * Checks if all edges have valid nodes
    * Checks that there are no duplicate edges
    * Checks if from/to is valid

    Returns True if valid, False otherwise.
    '''

    if not flow:
        flow = chip.get('option', 'flow')

    error = False

    nodes = set()
    for (step, index) in _get_flowgraph_nodes(chip, flow):
        nodes.add((step, index))
        input_nodes = chip.get('flowgraph', flow, step, index, 'input')
        nodes.update(input_nodes)

        for node in set(input_nodes):
            if input_nodes.count(node) > 1:
                in_step, in_index = node
                chip.logger.error(f'Duplicate edge from {in_step}{in_index} to '
                                  f'{step}{index} in the {flow} flowgraph')
                error = True

    for step, index in nodes:
        # For each task, check input requirements.
        tool, task = get_tool_task(chip, step, index, flow=flow)

        if not tool:
            chip.logger.error(f'{step}{index} is missing a tool definition in the {flow} '
                              'flowgraph')
            error = True

        if not task:
            chip.logger.error(f'{step}{index} is missing a task definition in the {flow} '
                              'flowgraph')
            error = True

    for step in chip.get('option', 'from'):
        if step not in chip.getkeys('flowgraph', flow):
            chip.logger.error(f'{step} is not defined in the {flow} flowgraph')
            error = True

    for step in chip.get('option', 'to'):
        if step not in chip.getkeys('flowgraph', flow):
            chip.logger.error(f'{step} is not defined in the {flow} flowgraph')
            error = True

    for step, index in chip.get('option', 'prune'):
        if step not in chip.getkeys('flowgraph', flow):
            chip.logger.error(f'{step} is not defined in the {flow} flowgraph')
            error = True
        elif str(index) not in chip.getkeys('flowgraph', flow, step):
            chip.logger.error(f'{step}{index} is not defined in the {flow} flowgraph')
            error = True

    unreachable_steps = _unreachable_steps_to_execute(chip, flow)
    if unreachable_steps:
        chip.logger.error(f'These final steps in {flow} can not be reached: '
                          f'{list(unreachable_steps)}')
        error = True

    return not error


###########################################################################
def _check_flowgraph_io(chip, nodes=None):
    '''Check if flowgraph is valid in terms of input and output files.

    Returns True if valid, False otherwise.
    '''
    flow = chip.get('option', 'flow')
    if not nodes:
        nodes = nodes_to_execute(chip)
    for (step, index) in nodes:
        # For each task, check input requirements.
        tool, task = get_tool_task(chip, step, index, flow=flow)

        if tool == 'builtin':
            # We can skip builtins since they don't have any particular
            # input requirements -- they just pass through what they
            # receive.
            continue

        # Get files we receive from input nodes.
        in_nodes = _get_flowgraph_node_inputs(chip, flow, (step, index))
        all_inputs = set()
        requirements = chip.get('tool', tool, 'task', task, 'input', step=step, index=index)
        for in_step, in_index in in_nodes:
            if (in_step, in_index) not in nodes:
                # If we're not running the input step, the required
                # inputs need to already be copied into the build
                # directory.
                workdir = chip.getworkdir(step=in_step, index=in_index)
                in_step_out_dir = os.path.join(workdir, 'outputs')

                if not os.path.isdir(in_step_out_dir):
                    # This means this step hasn't been run, but that
                    # will be flagged by a different check. No error
                    # message here since it would be redundant.
                    inputs = []
                    continue

                design = chip.get('design')
                manifest = f'{design}.pkg.json'
                inputs = [inp for inp in os.listdir(in_step_out_dir) if inp != manifest]
            else:
                inputs = _gather_outputs(chip, in_step, in_index)

            for inp in inputs:
                node_inp = input_file_node_name(inp, in_step, in_index)
                if node_inp in requirements:
                    inp = node_inp
                if inp in all_inputs:
                    chip.logger.error(f'Invalid flow: {step}{index} '
                                      f'receives {inp} from multiple input tasks')
                    return False
                all_inputs.add(inp)

        for requirement in requirements:
            if requirement not in all_inputs:
                chip.logger.error(f'Invalid flow: {step}{index} will '
                                  f'not receive required input {requirement}.')
                return False

    return True


###########################################################################
def _gather_outputs(chip, step, index):
    '''Return set of filenames that are guaranteed to be in outputs
    directory after a successful run of step/index.'''

    flow = chip.get('option', 'flow')
    task_gather = getattr(chip._get_task_module(step, index, flow=flow, error=False),
                          '_gather_outputs',
                          None)
    if task_gather:
        return set(task_gather(chip, step, index))

    tool, task = get_tool_task(chip, step, index, flow=flow)
    return set(chip.get('tool', tool, 'task', task, 'output', step=step, index=index))


def _get_flowgraph_information(chip, flow, io=True):
    from siliconcompiler.scheduler import _setup_node
    from siliconcompiler.tools._common import input_provides, input_file_node_name

    # Save schema to avoid making permanent changes
    org_schema = chip.schema
    chip.schema = chip.schema.copy()

    # Setup nodes
    node_exec_order = _get_flowgraph_execution_order(chip, flow)
    if io:
        # try:
        for layer_nodes in node_exec_order:
            for step, index in layer_nodes:
                _setup_node(chip, step, index, flow=flow)
        # except:  # noqa E722
        #     io = False

    node_rank = {}
    for rank, rank_nodes in enumerate(node_exec_order):
        for step, index in rank_nodes:
            node_rank[f'{step}{index}'] = rank

    graph_inputs = {}
    all_graph_inputs = set()
    if io:
        for step, index in _get_flowgraph_nodes(chip, flow):
            tool, task = get_tool_task(chip, step, index, flow=flow)
            for keypath in chip.get('tool', tool, 'task', task, 'require', step=step, index=index):
                key = tuple(keypath.split(','))
                if key[0] == 'input':
                    graph_inputs.setdefault((step, index), set()).add(keypath)

        for inputs in graph_inputs.values():
            all_graph_inputs.update(inputs)

    exit_nodes = [f'{step}{index}' for step, index in _get_flowgraph_exit_nodes(chip, flow)]

    nodes = {}
    edges = []

    def clean_label(label):
        return label.replace("<", "").replace(">", "")

    def clean_text(label):
        return label.replace("<", r"\<").replace(">", r"\>")

    all_nodes = [(step, index) for step, index in sorted(_get_flowgraph_nodes(chip, flow))
                 if chip.get('record', 'status', step=step, index=index) != NodeStatus.SKIPPED]
    for step, index in all_nodes:
        tool, task = get_tool_task(chip, step, index, flow=flow)

        if io:
            inputs = chip.get('tool', tool, 'task', task, 'input', step=step, index=index)
            outputs = chip.get('tool', tool, 'task', task, 'output', step=step, index=index)
            if chip.get('record', 'status', step=step, index=index) == NodeStatus.SKIPPED:
                continue
        else:
            inputs = []
            outputs = []

        node = f'{step}{index}'
        if io and (step, index) in graph_inputs:
            inputs.extend(graph_inputs[(step, index)])

        nodes[node] = {
            "node": (step, index),
            "file_inputs": inputs,
            "inputs": {clean_text(f): f'input-{clean_label(f)}' for f in sorted(inputs)},
            "outputs": {clean_text(f): f'output-{clean_label(f)}' for f in sorted(outputs)},
            "task": f'{tool}/{task}' if tool != 'builtin' else task,
            "is_input": node_rank[node] == 0,
            "rank": node_rank[node]
        }
        nodes[node]["width"] = max(len(nodes[node]["inputs"]), len(nodes[node]["outputs"]))

        if tool is None or task is None:
            nodes[node]["task"] = None

        rank_diff = {}
        for in_step, in_index in _get_flowgraph_node_inputs(chip, flow, (step, index)):
            rank_diff[f'{in_step}{in_index}'] = node_rank[node] - node_rank[f'{in_step}{in_index}']
        nodes[node]["rank_diff"] = rank_diff

    for step, index in all_nodes:
        node = f'{step}{index}'
        if io:
            # get inputs
            edge_stats = {}
            for infile, in_nodes in input_provides(chip, step, index, flow=flow).items():
                outfile = infile
                for in_step, in_index in in_nodes:
                    infile = outfile
                    if infile not in nodes[node]["file_inputs"]:
                        infile = input_file_node_name(infile, in_step, in_index)
                        if infile not in nodes[node]["file_inputs"]:
                            continue
                    in_node_name = f"{in_step}{in_index}"
                    outlabel = f"{in_node_name}:output-{clean_label(outfile)}"
                    inlabel = f"{step}{index}:input-{clean_label(infile)}"

                    if in_node_name not in edge_stats:
                        edge_stats[in_node_name] = {
                            "count": 0,
                            "pairs": [],
                            "weight": min(nodes[node]["width"], nodes[in_node_name]["width"])
                        }
                    edge_stats[in_node_name]["count"] += 1
                    edge_stats[in_node_name]["pairs"].append((outlabel, inlabel))

            # assign edge weights

            # scale multiple weights
            for edge_data in edge_stats.values():
                edge_data["weight"] = int(
                    math.floor(max(1, edge_data["weight"] / edge_data["count"])))

            # lower exit nodes weights
            if node in exit_nodes:
                for edge_data in edge_stats.values():
                    edge_data["weight"] = 1
            else:
                for edge_data in edge_stats.values():
                    edge_data["weight"] *= 2

            # adjust for rank differences, lower weight if rankdiff is greater than 1
            for in_node, edge_data in edge_stats.items():
                if nodes[node]["rank_diff"][in_node] > 1:
                    edge_data["weight"] = 1

            # create edges
            for edge_data in edge_stats.values():
                for outlabel, inlabel in edge_data["pairs"]:
                    edges.append([outlabel, inlabel, edge_data["weight"]])

            if (step, index) in graph_inputs:
                for key in graph_inputs[(step, index)]:
                    inlabel = f"{step}{index}:input-{clean_label(key)}"
                    edges.append((key, inlabel, 1))
        else:
            all_inputs = []
            for in_step, in_index in chip.get('flowgraph', flow, step, index, 'input'):
                all_inputs.append(f'{in_step}{in_index}')
            for item in all_inputs:
                edges.append((item, node, 1 if node in exit_nodes else 2))

    # Restore schema
    chip.schema = org_schema

    return all_graph_inputs, nodes, edges, io



import siliconcompiler
from siliconcompiler.utils import register_sc_data_source


####################################################
# Setup for ICE40 Family FPGAs
####################################################
def setup():
    '''
    Lattice ICE40 FPGAs are a family of small parts
    made by Lattice Semiconductor.  A fully open source
    RTL to bitstream flow for ICE40 is available using
    yosys + nextpnr
    '''

    vendor = 'lattice'

    lut_size = '4'

    all_fpgas = []

    all_part_names = [
        "ice40up5k-sg48",
    ]

    for part_name in all_part_names:
        fpga = siliconcompiler.FPGA(part_name, package='siliconcompiler_data')
        register_sc_data_source(fpga)

        fpga.set('fpga', part_name, 'vendor', vendor)
        fpga.set('fpga', part_name, 'lutsize', lut_size)

        all_fpgas.append(fpga)

    return all_fpgas


#########################
if __name__ == "__main__":
    for fpga in setup(siliconcompiler.Chip('<fpga>')):
        fpga.write_manifest(f'{fpga.design}.json')



import base64
import json
import os
import pandas

import streamlit
from streamlit_agraph import agraph
import streamlit_antd_components as sac

from PIL import Image

from siliconcompiler import __version__ as sc_version
from siliconcompiler import utils
from siliconcompiler.report import report

from siliconcompiler.report.dashboard import state
from siliconcompiler.report.dashboard import layouts
from siliconcompiler.report.dashboard.utils import file_utils
from siliconcompiler.report.dashboard.components import flowgraph


SC_ABOUT = [
    f"SiliconCompiler {sc_version}",
    '''A compiler framework that automates translation from source code to
     silicon.''',
    "https://www.siliconcompiler.com/",
    "https://github.com/siliconcompiler/siliconcompiler/"
]
SC_MENU = {
    "Get help": "https://docs.siliconcompiler.com/",
    "Report a Bug":
    '''https://github.com/siliconcompiler/siliconcompiler/issues''',
    "About": "\n\n".join(SC_ABOUT)}
SC_DATA_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', '..', 'data'))
SC_LOGO_PATH = os.path.join(SC_DATA_ROOT, 'logo.png')
SC_FONT_PATH = os.path.join(SC_DATA_ROOT, 'RobotoMono', 'RobotoMono-Regular.ttf')


def page_header(title_col_width=0.7):
    """
    Displays the title and a selectbox that allows you to select a given run
    to inspect.

    Args:
        title_col_width (float) : A number between 0 and 1 which is the percentage of the
            width of the screen given to the title and logo. The rest is given to selectbox.
    """

    if state.DEVELOPER:
        col_width = (1 - title_col_width) / 2
        title_col, job_select_col, settings_col = \
            streamlit.columns([title_col_width, col_width, col_width], gap="large")
    else:
        title_col, job_select_col = \
            streamlit.columns([title_col_width, 1 - title_col_width], gap="large")

    with title_col:
        design_title(design=state.get_chip().design)
    with job_select_col:
        job_selector()

    if state.DEVELOPER:
        with settings_col:
            with streamlit.popover("Settings", use_container_width=True):
                all_layouts = layouts.get_all_layouts()
                layout_index = all_layouts.index(state.get_key(state.APP_LAYOUT))
                if state.set_key(
                        state.APP_LAYOUT,
                        streamlit.selectbox("Layout", all_layouts, index=layout_index)):
                    state.set_key(state.APP_RERUN, "Layout")

                state._DEBUG = streamlit.checkbox("Debug", state._DEBUG)

                state.set_key(
                    state.APP_RUNNING_REFRESH,
                    streamlit.slider(
                        "Running refresh rate (ms)",
                        min_value=1000,
                        max_value=10000,
                        step=500,
                        value=state.get_key(state.APP_RUNNING_REFRESH)))

                state.set_key(
                    state.APP_STOPPED_REFRESH,
                    streamlit.slider(
                        "Stopped refresh rate (ms)",
                        min_value=1000,
                        max_value=100000,
                        step=1000,
                        value=state.get_key(state.APP_STOPPED_REFRESH)))

                state.set_key(
                    state.MAX_DICT_ITEMS_TO_SHOW,
                    streamlit.number_input(
                        "Maximum dict item to show",
                        min_value=1,
                        max_value=10000,
                        value=state.get_key(state.MAX_DICT_ITEMS_TO_SHOW)))

                state.set_key(
                    state.MAX_FILE_LINES_TO_SHOW,
                    streamlit.number_input(
                        "Maximum file lines to show",
                        min_value=100,
                        max_value=1000,
                        step=100,
                        value=state.get_key(state.MAX_FILE_LINES_TO_SHOW)))


def design_title(design=""):
    font = base64.b64encode(open(SC_FONT_PATH, "rb").read()).decode()
    streamlit.html(
        f'''
<head>
    <style>
        /* Define the @font-face rule */
        @font-face {{
        font-family: 'Roboto Mono';
        src: url(data:font/truetype;charset=utf-8;base64,{font}) format('truetype');
        font-weight: normal;
        font-style: normal;
        }}

        /* Styles for the logo and text */
        .logo-container {{
        display: flex;
        align-items: flex-start;
        }}

        .logo-image {{
        margin-right: 10px;
        margin-top: -10px;
        }}

        .logo-text {{
        display: flex;
        flex-direction: column;
        margin-top: -20px;
        }}

        .design-text {{
        color: #F1C437; /* Yellow color */
        font-family: 'Roboto Mono', sans-serif;
        font-weight: 700 !important;
        font-size: 30px !important;
        margin-bottom: -16px;
        }}

        .dashboard-text {{
        color: #1D4482; /* Blue color */
        font-family: 'Roboto Mono', sans-serif;
        font-weight: 700 !important;
        font-size: 30px !important;
        }}

    </style>
</head>
        '''
    )

    logo = base64.b64encode(open(SC_LOGO_PATH, "rb").read()).decode()
    streamlit.html(
        f'''
<body>
    <div class="logo-container">
        <img src="data:image/png;base64,{logo}" alt="SiliconCompiler logo"
             class="logo-image" height="61">
        <div class="logo-text">
            <p class="design-text">{design}</p>
            <p class="dashboard-text">dashboard</p>
        </div>
    </div>
</body>
        '''
    )


def job_selector():
    job = streamlit.selectbox(
        'pick a job',
        state.get_chips(),
        label_visibility='collapsed')

    if state.set_key(state.SELECTED_JOB, job):
        # Job changed, so need to run
        state.set_key(state.APP_RERUN, "Job")


def setup_page():
    streamlit.set_page_config(
        page_title=f'{state.get_chip().design} dashboard',
        page_icon=Image.open(SC_LOGO_PATH),
        layout="wide",
        menu_items=SC_MENU)


def file_viewer(chip, path, page_key=None, header_col_width=0.89):
    if not path:
        streamlit.error('Select a file')
        return

    if not os.path.isfile(path):
        streamlit.error(f'{path} is not a file')
        return

    # Detect file type
    relative_path = os.path.relpath(path, chip.getworkdir())
    filename = os.path.basename(path)
    file_extension = utils.get_file_ext(path)

    # Build streamlit module
    header_col, download_col = \
        streamlit.columns([header_col_width, 1 - header_col_width], gap='small')

    with header_col:
        streamlit.header(relative_path)

    with download_col:
        streamlit.markdown(' ')  # aligns download button with title
        streamlit.download_button(
            label="Download",
            data=path,
            file_name=filename,
            use_container_width=True)

    try:
        if file_extension in ('jpg', 'jpeg', 'png'):
            # Data is an image
            streamlit.image(path)
        elif file_extension == 'json':
            # Data is a json file
            data = json.loads(file_utils.read_file(path, None))
            expand_keys = report.get_total_manifest_key_count(data) < \
                state.get_key(state.MAX_DICT_ITEMS_TO_SHOW)
            if not expand_keys:
                # Open two levels
                expand_keys = 2
            streamlit.json(data, expanded=expand_keys)
        else:
            file_data = file_utils.read_file(path, None).splitlines()
            max_pages = len(file_data)
            page_size = state.get_key(state.MAX_FILE_LINES_TO_SHOW)

            file_section = streamlit.container()

            if page_key:
                if state.get_key(page_key) is None:
                    state.set_key(page_key, 1)
                index = state.get_key(page_key)
            else:
                index = 1

            page = sac.pagination(
                align='center',
                index=index,
                jump=True,
                show_total=True,
                page_size=page_size,
                total=max_pages,
                disabled=max_pages < state.get_key(state.MAX_FILE_LINES_TO_SHOW))

            if page_key:
                state.set_key(page_key, page)

            start_idx = (page - 1) * state.get_key(state.MAX_FILE_LINES_TO_SHOW)
            end_idx = start_idx + state.get_key(state.MAX_FILE_LINES_TO_SHOW)
            file_show = file_data[start_idx:end_idx]
            with file_section:
                # Assume file is text
                streamlit.code(
                    "\n".join(file_show),
                    language=file_utils.get_file_type(file_extension),
                    line_numbers=True)
    except Exception as e:
        streamlit.markdown(f'Error occurred reading file: {e}')


def manifest_viewer(
        chip,
        header_col_width=0.70):
    """
    Displays the manifest and a way to search through the manifest.

    Args:
        chip (Chip) : Chip object
        header_col_width (float) : A number between 0 and 1 which is the maximum
            percentage of the width of the screen given to the header. The rest
            is given to the settings and download buttons.
    """
    end_column_widths = (1 - header_col_width) / 2
    header_col, settings_col, download_col = \
        streamlit.columns(
            [header_col_width, end_column_widths, end_column_widths],
            gap='small')
    with header_col:
        streamlit.header('Manifest')

    with settings_col:
        streamlit.markdown(' ')  # aligns with title
        with streamlit.popover("Settings", use_container_width=True):
            if streamlit.checkbox(
                    'Raw manifest',
                    help='Click here to see the manifest before it was made more readable'):
                manifest_to_show = chip.schema.cfg
            else:
                manifest_to_show = report.make_manifest(chip)

            if streamlit.checkbox(
                    'Hide empty values',
                    help='Hide empty keypaths',
                    value=True):
                manifest_to_show = report.search_manifest(
                    manifest_to_show,
                    value_search='*')

            search_key = streamlit.text_input('Search Keys', '', placeholder="Keys")
            search_value = streamlit.text_input('Search Values', '', placeholder="Values")

            manifest_to_show = report.search_manifest(
                manifest_to_show,
                key_search=search_key,
                value_search=search_value)

    with download_col:
        streamlit.markdown(' ')  # aligns with title
        streamlit.download_button(
            label='Download',
            file_name='manifest.json',
            data=json.dumps(chip.schema.cfg, indent=2),
            mime="application/json",
            use_container_width=True)

    expand_keys = report.get_total_manifest_key_count(manifest_to_show) < \
        state.get_key(state.MAX_DICT_ITEMS_TO_SHOW)
    if not expand_keys:
        # Open two levels
        expand_keys = 2
    streamlit.json(manifest_to_show, expanded=expand_keys)


def metrics_viewer(metric_dataframe, metric_to_metric_unit_map, header_col_width=0.7, height=None):
    """
    Displays multi-select check box to the users which allows them to select
    which nodes and metrics to view in the dataframe.

    Args:
        metric_dataframe (Pandas.DataFrame) : Contains the metrics of all nodes.
        metric_to_metric_unit_map (dict) : Maps the metric to the associated metric unit.
    """

    all_nodes = metric_dataframe.columns.tolist()
    all_metrics = list(metric_to_metric_unit_map.values())

    header_col, settings_col = streamlit.columns(
        [header_col_width, 1 - header_col_width],
        gap="large")
    with header_col:
        streamlit.header('Metrics')
    with settings_col:
        # Align to header
        streamlit.markdown('')

        with streamlit.popover("Settings", use_container_width=True):
            transpose = streamlit.checkbox(
                'Transpose',
                help='Transpose the metrics table')

            display_nodes = streamlit.multiselect('Pick nodes to include', all_nodes, [])
            display_metrics = streamlit.multiselect('Pick metrics to include?', all_metrics, [])

    # Filter data
    if not display_nodes:
        display_nodes = all_nodes

    if not display_metrics:
        display_metrics = all_metrics

    dataframe_nodes = list(display_nodes)
    dataframe_metrics = []
    for metric in metric_dataframe.index.tolist():
        if metric_to_metric_unit_map[metric] in display_metrics:
            dataframe_metrics.append(metric)

    metric_dataframe = metric_dataframe.loc[dataframe_metrics, dataframe_nodes]
    if transpose:
        metric_dataframe = metric_dataframe.transpose()

    # TODO By July 2024, Streamlit will let catch click events on the dataframe
    streamlit.dataframe(metric_dataframe, use_container_width=True, height=height)


def node_image_viewer(chip, step, index):
    exts = ('png', 'jpg', 'jpeg')
    images = []
    for path, _, files in report.get_files(chip, step, index):
        images.extend([os.path.join(path, f) for f in files if utils.get_file_ext(f) in exts])

    if not images:
        streamlit.markdown("No images to show")

    work_dir = chip.getworkdir(step=step, index=index)

    columns = streamlit.slider(
        "Image columns",
        min_value=1,
        max_value=min(len(images), 10),
        value=min(len(images), 4),
        disabled=len(images) < 2)

    column = 0
    for image in sorted(images):
        if column == 0:
            cols = streamlit.columns(columns)

        with cols[column]:
            streamlit.image(
                image,
                caption=os.path.relpath(image, work_dir),
                use_column_width=True)

        column += 1
        if column == columns:
            column = 0


def node_file_tree_viewer(chip, step, index):
    logs_and_reports = file_utils.convert_filepaths_to_select_tree(
        report.get_files(chip, step, index))

    if not logs_and_reports:
        streamlit.markdown("No files to show")

    lookup = {}
    tree_items = []

    file_metrics = report.get_metrics_source(chip, step, index)
    work_dir = chip.getworkdir(step=step, index=index)

    def make_item(file):
        lookup[file['value']] = file['label']
        item = sac.TreeItem(
            file['value'],
            icon=file_utils.get_file_icon(file['value']),
            tag=[],
            children=[])

        check_file = os.path.relpath(file['value'], work_dir)
        if check_file in file_metrics:
            for metric in file_metrics[check_file]:
                if len(item.tag) < 5:
                    item.tag.append(sac.Tag(metric, color='green'))
                else:
                    item.tag.append(sac.Tag('metrics...', color='geekblue'))
                    break
            item.tooltip = "metrics: " + ", ".join(file_metrics[check_file])

        if 'children' in file:
            item.icon = 'folder'
            for child in file['children']:
                item.children.append(make_item(child))

        return item

    for file in logs_and_reports:
        tree_items.append(make_item(file))

    def format_label(value):
        return lookup[value]

    selected = sac.tree(
        items=tree_items,
        format_func=format_label,
        size='md',
        icon='table',
        open_all=True)

    if selected and os.path.isfile(selected):
        state.set_key(state.SELECTED_FILE, selected)
        state.set_key(state.SELECTED_FILE_PAGE, None)


def node_viewer(chip, step, index, metric_dataframe, height=None):
    metrics_col, records_col, logs_and_reports_col = streamlit.columns(3, gap='small')

    node_name = f'{step}{index}'

    with metrics_col:
        streamlit.subheader(f'{node_name} metrics')
        if node_name in metric_dataframe:
            streamlit.dataframe(
                metric_dataframe[node_name].dropna(),
                use_container_width=True,
                height=height)
    with records_col:
        streamlit.subheader(f'{step}{index} details')
        nodes = {}
        nodes[step + index] = report.get_flowgraph_nodes(chip, step, index)
        streamlit.dataframe(
            pandas.DataFrame.from_dict(nodes),
            use_container_width=True,
            height=height)
    with logs_and_reports_col:
        streamlit.subheader(f'{step}{index} files')
        node_file_tree_viewer(chip, step, index)


def flowgraph_viewer(chip):
    '''
    This function creates, displays, and returns the selected node of the flowgraph.

    Args:
        chip (Chip) : The chip object that contains the schema read from.
    '''

    nodes, edges = flowgraph.get_nodes_and_edges(chip)
    if state.set_key(state.SELECTED_FLOWGRAPH_NODE, agraph(
            nodes=nodes,
            edges=edges,
            config=flowgraph.get_graph_config())):
        if state.get_key(state.SELECTED_FLOWGRAPH_NODE):
            state.set_key(state.NODE_SOURCE, "flowgraph")


def node_selector(nodes):
    """
    Displays selectbox for nodes to show in the node information panel. Since
    both the flowgraph and selectbox show which node's information is
    displayed, the one clicked more recently will be displayed.

    Args:
        nodes (list) : Contains the metrics of all nodes.
    """
    prev_node = state.get_selected_node()

    with streamlit.popover("Select Node", use_container_width=True):
        # Preselect node
        idx = 0
        if prev_node:
            idx = nodes.index(prev_node)
        if state.set_key(
                state.SELECTED_SELECTOR_NODE,
                streamlit.selectbox(
                    'Pick a node to inspect',
                    nodes,
                    index=idx)):
            state.set_key(state.NODE_SOURCE, "selector")



import os
import streamlit

from siliconcompiler.report.dashboard import components
from siliconcompiler.report.dashboard.components import graph
from siliconcompiler.report.dashboard import state
from siliconcompiler.report.dashboard import utils
from siliconcompiler.report.dashboard.utils import file_utils
from siliconcompiler.report.dashboard.layouts import _common
import streamlit_antd_components as sac


def layout():
    chip = state.get_chip()
    metric_dataframe, node_to_step_index_map, metric_to_metric_unit_map = \
        utils.generate_metric_dataframe(chip)

    components.page_header()

    tab_headings = [
        sac.TabsItem(
            "Metrics",
            icon='stack'),
        sac.TabsItem(
            "Node Information",
            icon='diagram-2'),
        sac.TabsItem(
            "Manifest",
            icon=file_utils.get_file_icon('manifest.pkg.json')),
        sac.TabsItem(
            "File Viewer",
            icon=file_utils.get_file_icon(state.get_key(state.SELECTED_FILE))),
        sac.TabsItem(
            "Design Preview",
            icon=file_utils.get_file_icon('design.png'),
            disabled=not os.path.isfile(f'{chip.getworkdir()}/{chip.design}.png')),
        sac.TabsItem(
            "Graphs",
            icon='graph-up',
            disabled=len(state.get_key(state.LOADED_CHIPS)) == 1)
    ]

    tab_selected = _common.sac_tabs(tab_headings)

    if tab_selected == "Metrics":
        # Add flowgraph
        if state.get_key(state.DISPLAY_FLOWGRAPH):
            default_flowgraph_width_in_percent = 0.4
            flowgraph_col_width_in_pixels = 520
            flowgraph_col_width_in_percent = \
                state.compute_component_size(
                    default_flowgraph_width_in_percent,
                    flowgraph_col_width_in_pixels)

            flowgraph_col, metrics_container = \
                streamlit.columns(
                    [flowgraph_col_width_in_percent, 1 - flowgraph_col_width_in_percent],
                    gap="large")

            with flowgraph_col:
                header_col, flowgraph_toggle_container = streamlit.columns(2, gap="large")
                with header_col:
                    streamlit.header('Flowgraph')
                components.flowgraph_viewer(chip)
        else:
            flowgraph_toggle_container = streamlit.container()
            metrics_container = streamlit.container()

        with flowgraph_toggle_container:
            streamlit.markdown("")
            streamlit.markdown("")

            if state.set_key(state.DISPLAY_FLOWGRAPH, not streamlit.checkbox(
                    'Hide flowgraph',
                    help='Click here to hide the flowgraph')):
                state.set_key(state.APP_RERUN, "Flowgraph")

        with metrics_container:
            components.metrics_viewer(
                metric_dataframe,
                metric_to_metric_unit_map,
                height=1000)

    if tab_selected == "Node Information":
        header_col, settings_col = \
            streamlit.columns(
                [0.7, 0.3],
                gap='small')
        with header_col:
            streamlit.header('Node Information')
        with settings_col:
            components.node_selector(list(node_to_step_index_map.keys()))

        if state.get_selected_node():
            step, index = node_to_step_index_map[state.get_selected_node()]
            current_file = state.get_key(state.SELECTED_FILE)
            components.node_viewer(chip, step, index, metric_dataframe, height=1000)
            if state.get_key(state.SELECTED_FILE) and \
                    current_file != state.get_key(state.SELECTED_FILE):
                state.set_key(state.APP_RERUN, "File")

    if tab_selected == "Manifest":
        components.manifest_viewer(chip)

    if tab_selected == "File Viewer":
        components.file_viewer(
            chip,
            state.get_key(state.SELECTED_FILE),
            page_key=state.SELECTED_FILE_PAGE)

    if tab_selected == "Design Preview":
        components.file_viewer(chip, f'{chip.getworkdir()}/{chip.design}.png')

    if tab_selected == "Graphs":
        graph.viewer(node_to_step_index_map)

    _common.check_rerun()



from siliconcompiler.tools.yosys import syn_setup, syn_post_process
import json
from siliconcompiler import sc_open
from siliconcompiler.tools._common import get_tool_task, record_metric


######################################################################
# Make Docs
######################################################################
def make_docs(chip):
    from siliconcompiler.targets import fpgaflow_demo
    chip.set('fpga', 'partname', 'ice40up5k-sg48')
    chip.use(fpgaflow_demo)


def setup(chip):
    '''
    Perform FPGA synthesis
    '''

    # Generic synthesis task setup.
    syn_setup(chip)

    # FPGA-specific setup.
    setup_fpga(chip)


def setup_fpga(chip):
    ''' Helper method for configs specific to FPGA steps (both syn and lec).
    '''

    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)
    design = chip.top()

    part_name = chip.get('fpga', 'partname')

    # Require that a lut size is set for FPGA scripts.
    chip.add('tool', tool, 'task', task, 'require',
             ",".join(['fpga', part_name, 'lutsize']),
             step=step, index=index)

    if chip.valid('fpga', part_name, 'file', 'yosys_flop_techmap') and \
       chip.get('fpga', part_name, 'file', 'yosys_flop_techmap'):

        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['fpga', part_name, 'file', 'yosys_flop_techmap']),
                 step=step, index=index)

    if chip.valid('fpga', part_name, 'file', 'yosys_dsp_techmap') and \
       chip.get('fpga', part_name, 'file', 'yosys_dsp_techmap'):

        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['fpga', part_name, 'file', 'yosys_dsp_techmap']),
                 step=step, index=index)

    if chip.valid('fpga', part_name, 'file', 'yosys_extractlib') and \
       chip.get('fpga', part_name, 'file', 'yosys_extractlib'):

        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['fpga', part_name, 'file', 'yosys_extractlib']),
                 step=step, index=index)

    if chip.valid('fpga', part_name, 'file', 'yosys_macrolib') and \
       chip.get('fpga', part_name, 'file', 'yosys_macrolib'):

        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['fpga', part_name, 'file', 'yosys_macrolib']),
                 step=step, index=index)

    part_name = chip.get('fpga', 'partname')
    for resource in ('yosys_registers', 'yosys_brams', 'yosys_dsps'):
        if not chip.valid('fpga', part_name, 'var', resource):
            continue
        if not chip.get('fpga', part_name, 'var', resource):
            continue
        chip.add('tool', tool, 'task', task, 'require', f'fpga,{part_name},var,{resource}',
                 step=step, index=index)

    # Verify memory techmapping setup.  If a memory libmap
    # is provided a memory techmap verilog file is needed too
    if (chip.valid('fpga', part_name, 'file', 'yosys_memory_libmap') and
        chip.get('fpga', part_name, 'file', 'yosys_memory_libmap')) or \
        (chip.valid('fpga', part_name, 'file', 'yosys_memory_techmap') and
         chip.get('fpga', part_name, 'file', 'yosys_memory_techmap')):

        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['fpga', part_name, 'file', 'yosys_memory_libmap']),
                 step=step, index=index)
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['fpga', part_name, 'file', 'yosys_memory_techmap']),
                 step=step, index=index)

    chip.add('tool', tool, 'task', task, 'output', design + '.blif', step=step, index=index)


##################################################
def post_process(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    part_name = chip.get('fpga', 'partname')

    syn_post_process(chip)

    with sc_open("reports/stat.json") as f:
        metrics = json.load(f)
        if "design" in metrics:
            metrics = metrics["design"]
        else:
            return

        if "num_cells_by_type" in metrics:
            metrics = metrics["num_cells_by_type"]
        else:
            return

        dff_cells = []
        if chip.valid('fpga', part_name, 'var', 'yosys_registers'):
            dff_cells = chip.get('fpga', part_name, 'var', 'yosys_registers')
        brams_cells = []
        if chip.valid('fpga', part_name, 'var', 'yosys_brams'):
            brams_cells = chip.get('fpga', part_name, 'var', 'yosys_brams')
        dsps_cells = []
        if chip.valid('fpga', part_name, 'var', 'yosys_dsps'):
            dsps_cells = chip.get('fpga', part_name, 'var', 'yosys_dsps')

        data = {
            "registers": 0,
            "luts": 0,
            "dsps": 0,
            "brams": 0
        }
        for cell, count in metrics.items():
            if cell == "$lut":
                data["luts"] += count
            elif cell in dff_cells:
                data["registers"] += count
            elif cell in dsps_cells:
                data["dsps"] += count
            elif cell in brams_cells:
                data["brams"] += count

        for metric, value in data.items():
            record_metric(chip, step, index, metric, value, "reports/stat.json")



import re
from siliconcompiler.utils import sc_open
from siliconcompiler.tools._common.asic import get_tool_task


def __get_clock_data(chip, clock_units_multiplier=1):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    period = None
    # get clock information from sdc files
    if chip.valid('input', 'constraint', 'sdc'):
        for sdc in chip.find_files('input', 'constraint', 'sdc', step=step, index=index):
            lines = []
            with sc_open(sdc) as f:
                lines = f.read().splitlines()

            # collect simple variables in case clock is specified with a variable
            re_var = r"[A-Za-z0-9_]+"
            re_num = r"[0-9\.]+"
            sdc_vars = {}
            for line in lines:
                tcl_variable = re.findall(fr"^\s*set\s+({re_var})\s+({re_num}|\${re_var})", line)
                if tcl_variable:
                    var_name, var_value = tcl_variable[0]
                    sdc_vars[f'${var_name}'] = var_value

            # TODO: handle line continuations
            for line in lines:
                clock_period = re.findall(fr"create_clock\s.*-period\s+({re_num}|\${re_var})",
                                          line)
                if clock_period:
                    convert_period = clock_period[0]
                    while isinstance(convert_period, str) and convert_period[0] == "$":
                        if convert_period in sdc_vars:
                            convert_period = sdc_vars[convert_period]
                        else:
                            break
                    if isinstance(convert_period, str) and convert_period[0] == "$":
                        chip.logger.warning('Unable to identify clock period from '
                                            f'{clock_period[0]}.')
                        continue
                    else:
                        try:
                            clock_period = float(convert_period)
                        except TypeError:
                            continue

                    clock_period = clock_period * clock_units_multiplier

                    if period is None:
                        period = clock_period
                    else:
                        period = min(period, clock_period)

        if period is not None:
            return period, None, [('input', 'constraint', 'sdc')]

    if period is None:
        keys = []
        key_pin = None
        # get clock information from defined clocks
        for pin in chip.getkeys('datasheet', 'pin'):
            for mode in chip.getkeys('datasheet', 'pin', pin, 'type'):
                if chip.get('datasheet', 'pin', pin, 'type', mode) == 'clock':
                    clock_period = min(chip.get('datasheet', 'pin', pin, 'tperiod', mode)) * 1e9

                    if period is None:
                        period = clock_period
                        keys = [
                            ('datasheet', 'pin', pin, 'type', mode),
                            ('datasheet', 'pin', pin, 'tperiod', mode)
                        ]
                        key_pin = pin
                    else:
                        if clock_period < period:
                            period = clock_period
                            keys = [
                                ('datasheet', 'pin', pin, 'type', mode),
                                ('datasheet', 'pin', pin, 'tperiod', mode)
                            ]
                            key_pin = pin
        return period, key_pin, keys

    return None, None, []


def add_clock_requirements(chip):
    _, _, keys = __get_clock_data(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)
    for key in keys:
        chip.add('tool', tool, 'task', task, 'require', ','.join(key),
                 step=step, index=index)


def get_clock_period(chip, clock_units_multiplier=1):
    period, name, _ = __get_clock_data(chip, clock_units_multiplier=clock_units_multiplier)
    return name, period



from siliconcompiler.tools.montage import montage
from siliconcompiler.tools._common import get_tool_task


def setup(chip):
    '''
    Tiles input images into a single output image.

    Notes:
    Need to make ensure that /etc/ImageMagick-6/policy.xml
      <policy domain="resource" name="memory" value="8GiB"/>
      <policy domain="resource" name="map" value="8GiB"/>
      <policy domain="resource" name="width" value="32KP"/>
      <policy domain="resource" name="height" value="32KP"/>
      <policy domain="resource" name="area" value="1GP"/>
      <policy domain="resource" name="disk" value="8GiB"/>
    This ensures there are enough resources available to generate
    the final image.
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)
    design = chip.top()

    montage.setup(chip)

    chip.set('tool', tool, 'task', task, 'var', 'xbins', '2',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'ybins', '2',
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'var', 'xbins',
             'Number of bins along the x-axis',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'ybins',
             'Number of bins along the y-axis',
             field='help')

    xbins = int(chip.get('tool', tool, 'task', task, 'var', 'xbins',
                         step=step, index=index)[0])
    ybins = int(chip.get('tool', tool, 'task', task, 'var', 'ybins',
                         step=step, index=index)[0])

    for x in range(xbins):
        for y in range(ybins):
            chip.add('tool', tool, 'task', task, 'input', f'{design}_X{x}_Y{y}.png',
                     step=step, index=index)

    chip.set('tool', tool, 'task', task, 'output', f'{design}.png',
             step=step, index=index)

    options = []

    for y in range(ybins):
        for x in range(xbins):
            options.append(f'inputs/{design}_X{x}_Y{y}.png')

    options.append('-tile')
    options.append(f'{xbins}x{ybins}')
    options.append('-geometry')
    options.append('+0+0')
    options.append(f'outputs/{design}.png')

    chip.set('tool', tool, 'task', task, 'option', options,
             step=step, index=index)



from siliconcompiler.tools.verilator.verilator import setup as setup_tool
from siliconcompiler.tools.verilator.verilator import runtime_options as runtime_options_tool
from siliconcompiler.tools._common import get_input_files, get_tool_task


def setup(chip):
    '''
    Compiles Verilog and C/C++ sources into an executable. In addition to the
    standard RTL inputs, this task reads C/C++ sources from :keypath:`input,
    hll, c`.  Outputs an executable in ``outputs/<design>.vexe``.
    '''

    # Generic tool setup.
    setup_tool(chip)

    tool = 'verilator'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    design = chip.top()

    chip.add('tool', tool, 'task', task, 'require', 'input,hll,c', step=step, index=index)
    chip.set('tool', tool, 'task', task, 'output', f'{design}.vexe', step=step, index=index)

    # var defaults
    chip.set('tool', tool, 'task', task, 'var', 'mode', 'cc', clobber=False, step=step, index=index)
    chip.set('tool', tool, 'task', task, 'var', 'trace', False, clobber=False,
             step=step, index=index)
    chip.set('tool', tool, 'task', task, 'var', 'trace_type', 'vcd', clobber=False,
             step=step, index=index)

    mode = chip.get('tool', tool, 'task', task, 'var', 'mode', step=step, index=index)
    if mode not in (['cc'], ['systemc']):
        chip.error(f"Invalid mode {mode} provided to verilator/compile. Expected one of 'cc' or "
                   "'systemc'")
    chip.add('tool', tool, 'task', task, 'require',
             ','.join(['tool', tool, 'task', task, 'var', 'mode']),
             step=step, index=index)

    trace_type = chip.get('tool', tool, 'task', task, 'var', 'trace_type', step=step, index=index)
    if trace_type not in (['vcd'], ['fst']):
        chip.error(f"Invalid trace type {trace_type} provided to verilator/compile. Expected "
                   "one of 'vcd' or 'fst'.")
    chip.add('tool', tool, 'task', task, 'require',
             ','.join(['tool', tool, 'task', task, 'var', 'trace_type']),
             step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'cflags',
             'flags to provide to the C++ compiler invoked by Verilator',
             field='help')
    if chip.get('tool', tool, 'task', task, 'var', 'cflags', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['tool', tool, 'task', task, 'var', 'cflags']),
                 step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'ldflags',
             'flags to provide to the linker invoked by Verilator',
             field='help')
    if chip.get('tool', tool, 'task', task, 'var', 'ldflags', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['tool', tool, 'task', task, 'var', 'ldflags']),
                 step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'pins_bv',
             'controls datatypes used to represent SystemC inputs/outputs. See --pins-bv in '
             'Verilator docs for more info.',
             field='help')
    if chip.get('tool', tool, 'task', task, 'var', 'pins_bv', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['tool', tool, 'task', task, 'var', 'pins_bv']),
                 step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'mode',
             "defines compilation mode for Verilator. Valid options are 'cc' for C++, or 'systemc' "
             "for SystemC.",
             field='help')

    chip.set('tool', tool, 'task', task, 'dir', 'cincludes',
             'include directories to provide to the C++ compiler invoked by Verilator',
             field='help')
    if chip.get('tool', tool, 'task', task, 'dir', 'cincludes', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['tool', tool, 'task', task, 'dir', 'cincludes']),
                 step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'trace',
             "if true, enables trace generation.",
             field='help')
    if chip.get('tool', tool, 'task', task, 'var', 'trace', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['tool', tool, 'task', task, 'var', 'trace']),
                 step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'trace_type',
             "specifies type of wave file to create when [trace] is set. Valid options are "
             "'vcd' or 'fst'. Defaults to 'vcd'.",
             field='help')

    chip.set('tool', tool, 'task', task, 'var', 'initialize_random',
             'true/false, when true registers will reset with a random value.',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'initialize_random', False,
             step=step, index=index, clobber=False)


def runtime_options(chip):
    tool = 'verilator'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)
    design = chip.top()

    cmdlist = runtime_options_tool(chip)

    random_init = chip.get('tool', tool, 'task', task, 'var',
                           'initialize_random', step=step, index=index)
    if random_init == ['true']:
        cmdlist.extend(['--x-assign', 'unique'])

    cmdlist.extend(['--exe', '--build'])

    threads = chip.get('tool', tool, 'task', task, 'threads', step=step, index=index)
    cmdlist.extend(['-j', str(threads)])

    mode = chip.get('tool', tool, 'task', task, 'var', 'mode', step=step, index=index)
    if mode == ['cc']:
        cmdlist.append('--cc')
    elif mode == ['systemc']:
        cmdlist.append('--sc')

    pins_bv = chip.get('tool', tool, 'task', task, 'var', 'pins_bv', step=step, index=index)
    if pins_bv:
        cmdlist.extend(['--pins-bv', pins_bv[0]])

    cmdlist.extend(['-o', f'../outputs/{design}.vexe'])

    c_flags = chip.get('tool', tool, 'task', task, 'var', 'cflags', step=step, index=index)
    c_includes = chip.find_files('tool', tool, 'task', task, 'dir', 'cincludes',
                                 step=step, index=index)

    if chip.get('tool', tool, 'task', task, 'var', 'trace', step=step, index=index)[0] == 'true':
        trace_type = chip.get('tool', tool, 'task', task, 'var', 'trace_type',
                              step=step, index=index)

        if trace_type == ['vcd']:
            ext = 'vcd'
            trace_opt = '--trace'
        elif trace_type == ['fst']:
            ext = 'fst'
            trace_opt = '--trace-fst'

        cmdlist.append(trace_opt)

        # add siliconcompiler specific defines
        c_flags.append("-DSILICONCOMPILER_TRACE_DIR=\\\"reports\\\"")
        c_flags.append(f"-DSILICONCOMPILER_TRACE_FILE=\\\"reports/{design}.{ext}\\\"")

    if c_includes:
        c_flags.extend([f'-I{include}' for include in c_includes])

    if c_flags:
        cflags_str = ' '.join(c_flags)
        cmdlist.extend(['-CFLAGS', f'"{cflags_str}"'])

    ld_flags = chip.get('tool', tool, 'task', task, 'var', 'ldflags', step=step, index=index)
    if ld_flags:
        ldflags_str = ' '.join(ld_flags)
        cmdlist.extend(['-LDFLAGS', f'"{ldflags_str}"'])

    for value in get_input_files(chip, 'input', 'hll', 'c'):
        cmdlist.append(value)

    return cmdlist



from siliconcompiler.tools._common import get_tool_task
from siliconcompiler.tools.openroad import make_docs as or_make_docs
from siliconcompiler.tools.openroad import show
from siliconcompiler.tools.openroad._apr import set_reports


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    or_make_docs(chip)
    chip.set('tool', 'openroad', 'task', 'screenshot', 'var', 'show_filepath', '<path>')


def setup(chip):
    '''
    Generate a PNG file from a layout file
    '''
    show.generic_show_setup(chip, True)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.add('tool', tool, 'task', task, 'output', f'{chip.top()}.png', step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'show_vertical_resolution', '1024',
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'var', 'include_report_images', 'false',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'include_report_images',
             'true/false, include the images in reports/',
             field='help')

    set_reports(chip, [
        # Images
        'placement_density',
        'routing_congestion',
        'power_density',
        'ir_drop',
        'clock_placement',
        'clock_trees',
        'optimization_placement'
    ])


def pre_process(chip):
    show.pre_process(chip)



import os
import shutil
from siliconcompiler.tools._common import \
    add_require_input, add_frontend_requires, get_frontend_options, get_input_files, \
    get_tool_task, has_input_files
from siliconcompiler import utils
from siliconcompiler import sc_open

# Directory inside step/index dir to store bsc intermediate results.
VLOG_DIR = 'verilog'
BSC_DIR = 'bluespec'


def setup(chip):
    '''
    Performs high level synthesis to generate a verilog output
    '''

    if not has_input_files(chip, 'input', 'hll', 'bsv'):
        return "no files in [input,hll,bsv]"

    tool = 'bluespec'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    # Standard Setup
    refdir = 'tools/' + tool
    chip.set('tool', tool, 'exe', 'bsc')
    # This is technically the 'verbose' flag, but used alone it happens to give
    # us the version and exit cleanly, so we'll use it here.
    chip.set('tool', tool, 'vswitch', '-v')
    chip.set('tool', tool, 'version', '>=2021.07', clobber=False)

    chip.set('tool', tool, 'task', task, 'refdir', refdir,
             step=step, index=index,
             package='siliconcompiler', clobber=False)
    chip.set('tool', tool, 'task', task, 'threads', utils.get_cores(chip),
             step=step, index=index, clobber=False)

    # Input/Output requirements
    chip.add('tool', tool, 'task', task, 'output', chip.top() + '.v', step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', chip.top() + '.dot', step=step, index=index)

    # Schema requirements
    add_require_input(chip, 'input', 'hll', 'bsv')
    add_frontend_requires(chip, ['idir', 'ydir', 'define'])


################################
# Pre-process
################################
def pre_process(chip):
    # bsc requires its output directory exists before being called.
    for path in (VLOG_DIR, BSC_DIR):
        if os.path.isdir(path):
            shutil.rmtree(path)
        os.makedirs(path)


################################
#  Custom runtime options
################################
def runtime_options(chip):
    cmdlist = []

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    opts = get_frontend_options(chip, ['idir', 'ydir', 'define'])

    cmdlist.append('-verilog')
    cmdlist.append(f'-vdir {VLOG_DIR}')
    cmdlist.append(f'-bdir {BSC_DIR}')
    cmdlist.append('-info-dir reports')
    cmdlist.append('-u')
    cmdlist.append('-v')

    cmdlist.append('-show-module-use')
    cmdlist.append('-sched-dot')

    cmdlist.append(f'-g {chip.top(step, index)}')

    bsc_path = ':'.join(opts['ydir'] + ['%/Libraries'])
    cmdlist.append('-p ' + bsc_path)

    for value in opts['idir']:
        cmdlist.append('-I ' + value)
    for value in opts['define']:
        cmdlist.append('-D ' + value)

    sources = get_input_files(chip, 'input', 'hll', 'bsv', add_library_files=False)
    if len(sources) != 1:
        raise ValueError('Bluespec frontend only supports one source file!')
    cmdlist.extend(sources)

    return cmdlist


################################
# Post-process (post executable)
################################
def post_process(chip):
    ''' Tool specific function to run after step execution
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    shutil.copyfile(f"reports/{chip.top(step, index)}_combined_full.dot",
                    f"outputs/{chip.top()}.dot")

    extra_modules = set()
    use_file = os.path.join(VLOG_DIR, f"{chip.top(step, index)}.use")
    if os.path.exists(use_file):
        BSC_BASE = os.path.dirname(
            os.path.dirname(
                chip.get('record', 'toolpath', step=step, index=index)))
        BSC_LIB = os.path.join(BSC_BASE, "lib", "Verilog")

        with sc_open(use_file) as f:
            for module in f:
                module = module.strip()
                mod_path = os.path.join(BSC_LIB, f"{module}.v")
                if os.path.exists(mod_path):
                    extra_modules.add(mod_path)
                else:
                    chip.logger.warn(f"Unable to find module {module} source files at: {BSC_LIB}")

    # bsc outputs each compiled module to its own Verilog file, so we
    # concatenate them all to create a pickled output we can pass along.
    design = chip.top()
    with open(os.path.join('outputs', f'{design}.v'), 'w') as pickled_vlog:
        for src in os.listdir(VLOG_DIR):
            if src.endswith(".v"):
                with sc_open(os.path.join(VLOG_DIR, src)) as vlog_mod:
                    pickled_vlog.write(vlog_mod.read())

        pickled_vlog.write("\n")
        pickled_vlog.write("// Bluespec imports\n\n")
        for vfile in extra_modules:
            with sc_open(os.path.join(BSC_LIB, vfile)) as vlog_mod:
                pickled_vlog.write(vlog_mod.read() + "\n")



from siliconcompiler.tools._common import input_provides, add_common_file, get_tool_task
from siliconcompiler.tools._common.asic import set_tool_task_var, get_mainlib

from siliconcompiler.tools.openroad._apr import setup as apr_setup
from siliconcompiler.tools.openroad._apr import set_reports, set_pnr_inputs, set_pnr_outputs
from siliconcompiler.tools.openroad._apr import \
    define_ord_params, define_sta_params, define_sdc_params, \
    define_tiecell_params, define_pad_params, define_ppl_params
from siliconcompiler.tools.openroad._apr import build_pex_corners, define_ord_files
from siliconcompiler.tools.openroad._apr import extract_metrics


def setup(chip):
    '''
    Perform floorplanning and initial pin placements
    '''

    # Generic apr tool setup.
    apr_setup(chip)

    # Task setup
    design = chip.top()
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'script', 'apr/sc_init_floorplan.tcl',
             step=step, index=index)

    # Setup task IO
    set_pnr_inputs(chip)
    set_pnr_outputs(chip)
    add_common_file(chip, 'sc_pin_constraint', 'tcl/sc_pin_constraints.tcl')

    # set default values for task
    define_ord_params(chip)
    define_sta_params(chip)
    define_sdc_params(chip)
    define_pad_params(chip)
    define_ppl_params(chip)
    define_tiecell_params(chip)

    set_tool_task_var(chip, param_key='ifp_snap_strategy',
                      default_value='site',
                      schelp='Snapping strategy to use when placing macros. '
                             'Allowed values: none, site, manufacturing_grid')

    set_tool_task_var(chip, param_key='remove_synth_buffers',
                      default_value=True,
                      schelp='remove buffers inserted by synthesis')

    set_tool_task_var(chip, param_key='remove_dead_logic',
                      default_value=True,
                      schelp='remove logic which does not drive a primary output')

    # Handle additional input files
    if chip.valid('input', 'asic', 'floorplan') and \
       chip.get('input', 'asic', 'floorplan', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['input', 'asic', 'floorplan']),
                 step=step, index=index)

    if f'{design}.vg' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', f'{design}.vg',
                 step=step, index=index)
    else:
        chip.add('tool', tool, 'task', task, 'require', 'input,netlist,verilog',
                 step=step, index=index)

    set_reports(chip, [
        'check_setup',
        'setup',
        'unconstrained',
        'power'
    ])

    mainlib = get_mainlib(chip)

    # Setup required
    for component in chip.getkeys('constraint', 'component'):
        for key in chip.getkeys('constraint', 'component', component):
            if chip.get('constraint', 'component', component, key, step=step, index=index):
                chip.add('tool', tool, 'task', task, 'require',
                         ','.join(['constraint', 'component', component, key]),
                         step=step, index=index)
    for pin in chip.getkeys('constraint', 'pin'):
        for key in chip.getkeys('constraint', 'pin', pin):
            if chip.get('constraint', 'pin', pin, key, step=step, index=index):
                chip.add('tool', tool, 'task', task, 'require',
                         ','.join(['constraint', 'pin', pin, key]),
                         step=step, index=index)
    for ifp in ('aspectratio', 'density', 'corearea', 'coremargin', 'outline'):
        if chip.get('constraint', ifp, step=step, index=index):
            chip.add('tool', tool, 'task', task, 'require',
                     ','.join(['constraint', ifp]),
                     step=step, index=index)
    if chip.valid('library', mainlib, 'option', 'file', 'openroad_tracks'):
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['library', mainlib, 'option', 'file', 'openroad_tracks']),
                 step=step, index=index)


def pre_process(chip):
    build_pex_corners(chip)
    define_ord_files(chip)


def post_process(chip):
    extract_metrics(chip)



from siliconcompiler import NodeStatus
from siliconcompiler import units
from siliconcompiler.flowgraph import _get_flowgraph_execution_order, _get_flowgraph_exit_nodes, \
    nodes_to_execute
from siliconcompiler.tools._common import get_tool_task


def _find_summary_image(chip, ext='png'):
    for nodes in reversed(_get_flowgraph_execution_order(chip, chip.get('option', 'flow'))):
        for step, index in nodes:
            layout_img = chip.find_result(ext, step=step, index=index)
            if layout_img:
                return layout_img
    return None


def _find_summary_metrics(chip, metrics_map):
    metrics = {}
    for nodes in reversed(_get_flowgraph_execution_order(chip, chip.get('option', 'flow'))):
        for step, index in nodes:
            for name, metric_info in metrics_map.items():
                if name in metrics:
                    continue

                metric, formatter = metric_info

                data = chip.get('metric', metric, step=step, index=index)
                if data is not None:
                    unit = None
                    if chip.schema.has_field('metric', metric, 'unit'):
                        unit = chip.get('metric', metric, field='unit')
                    if formatter:
                        metrics[name] = formatter(data, unit)
                    else:
                        metrics[name] = str(data)

    return metrics


def _collect_data(chip, flow=None, flowgraph_nodes=None, format_as_string=True):
    if not flow:
        flow = chip.get('option', 'flow')
    if not flowgraph_nodes:
        flowgraph_nodes = nodes_to_execute(chip)
        # only report tool based steps functions
        for (step, index) in list(flowgraph_nodes):
            tool, task = get_tool_task(chip, step, '0', flow=flow)
            if tool == 'builtin':
                index = flowgraph_nodes.index((step, index))
                del flowgraph_nodes[index]
    flowgraph_nodes = [
        node for node in flowgraph_nodes
        if chip.get('record', 'status', step=node[0], index=node[1]) != NodeStatus.SKIPPED
    ]

    # Collections for data
    nodes = []
    errors = {}
    metrics = {}
    metrics_unit = {}
    reports = {}

    # Build ordered list of nodes in flowgraph
    for level_nodes in _get_flowgraph_execution_order(chip, flow):
        nodes.extend(sorted(level_nodes))
    nodes = [node for node in nodes if node in flowgraph_nodes]
    for (step, index) in nodes:
        metrics[step, index] = {}
        reports[step, index] = {}

    # Gather data and determine which metrics to show
    # We show a metric if:
    # - at least one step in the steps has a non-zero weight for the metric -OR -
    #   at least one step in the steps set a value for it
    metrics_to_show = []
    for metric in chip.getkeys('metric'):

        # Get the unit associated with the metric
        metric_unit = None
        if chip.schema.has_field('metric', metric, 'unit'):
            metric_unit = chip.get('metric', metric, field='unit')
        metric_type = chip.get('metric', metric, field='type')

        show_metric = False
        for step, index in nodes:
            if metric in chip.getkeys('flowgraph', flow,
                                      step, index, 'weight') and \
               chip.get('flowgraph', flow, step, index, 'weight', metric):
                show_metric = True

            value = chip.get('metric', metric, step=step, index=index)
            if value is not None:
                show_metric = True
            tool, task = get_tool_task(chip, step, index, flow=flow)
            rpts = chip.get('tool', tool, 'task', task, 'report', metric,
                            step=step, index=index)

            errors[step, index] = chip.get('record', 'status', step=step, index=index) == \
                NodeStatus.ERROR

            if value is not None:
                value = _format_value(metric, value, metric_unit, metric_type, format_as_string)

            metrics[step, index][metric] = value
            reports[step, index][metric] = rpts

        if show_metric:
            metrics_to_show.append(metric)
            metrics_unit[metric] = metric_unit if metric_unit else ''

    if 'totaltime' in metrics_to_show:
        if not any([chip.get('metric', 'totaltime', step=node[0], index=node[1]) is None
                    for node in nodes]):
            nodes.sort(
                key=lambda node: chip.get('metric', 'totaltime', step=node[0], index=node[1]))

    return nodes, errors, metrics, metrics_unit, metrics_to_show, reports


def _format_value(metric, value, metric_unit, metric_type, format_as_string):
    if metric == 'memory':
        if format_as_string:
            return units.format_binary(value, metric_unit)
        value, metric = units.scale_binary(value, metric_unit)
    elif metric in ['exetime', 'tasktime', 'totaltime']:
        if format_as_string:
            return units.format_time(value)
    elif metric_type == 'int':
        if format_as_string:
            return str(value)
    else:
        if format_as_string:
            return units.format_si(value, metric_unit)
        value, metric = units.scale_si(value, metric_unit)
    return value


def _get_flowgraph_path(chip, flow, nodes_to_execute, only_include_successful=False):
    selected_nodes = set()
    to_search = []
    # Start search with any successful leaf nodes.
    flowgraph_steps = list(map(lambda node: node[0], nodes_to_execute))
    end_nodes = _get_flowgraph_exit_nodes(chip, flow, steps=flowgraph_steps)
    for node in end_nodes:
        if only_include_successful:
            if NodeStatus.is_success(chip.get('record', 'status', step=node[0], index=node[1])):
                selected_nodes.add(node)
                to_search.append(node)
        else:
            selected_nodes.add(node)
            to_search.append(node)
    # Search backwards, saving anything that was selected by leaf nodes.
    while len(to_search) > 0:
        node = to_search.pop(-1)
        input_nodes = chip.get('record', 'inputnode', step=node[0], index=node[1])
        for selected in input_nodes:
            if selected not in selected_nodes:
                selected_nodes.add(selected)
                to_search.append(selected)

    return [node for node in selected_nodes
            if chip.get('record', 'status', step=node[0], index=node[1]) != NodeStatus.SKIPPED]



import os
import pkgutil


def __get_library_keys(chip, include_asic=True, library=None, libraries=None):
    '''
    Returns a list of libraries included in this step/index

    Args:
        chip (Chip): Chip object
        include_asic (bool): include the asic libraries.
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    libs = []

    if not libraries:
        libraries = set()

    pref_key = []
    if library:
        pref_key = ['library', library]

    def get_libs(*key):
        if chip.valid(*key) and chip.get(*key, step=step, index=index):
            if chip.get(*key, step=step, index=index):
                return [key]
        return []

    if include_asic:
        libs.extend(get_libs(*pref_key, 'asic', 'logiclib'))
        libs.extend(get_libs(*pref_key, 'asic', 'macrolib'))

    libnames = set()
    if library:
        libnames.add(library)
    for lib_key in get_libs(*pref_key, 'option', 'library'):
        if lib_key in libs:
            continue

        libs.append(lib_key)

        for libname in chip.get(*lib_key, step=step, index=index):
            if libname in libnames:
                continue

            libnames.add(libname)
            libs.extend(__get_library_keys(
                chip,
                include_asic=include_asic,
                library=libname,
                libraries=libnames))

    return set(libs)


def get_libraries(chip, include_asic=True):
    '''
    Returns a list of libraries included in this step/index

    Args:
        chip (Chip): Chip object
        include_asic (bool): include the asic libraries.
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    libs = []

    for key in __get_library_keys(chip, include_asic=include_asic):
        libs.extend(chip.get(*key, step=step, index=index))

    return set(libs)


def add_require_input(chip, *key, include_library_files=True):
    '''
    Adds input files to the require list of the task.

    Args:
        chip (Chip): Chip object
        key (list): Key to check for requirements
        add_library_files (bool): When True, files from library keys
            will be included
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    keys = []
    for key in __get_keys(chip, *key, include_library_files=False):
        keys.append(key)

    if include_library_files:
        for item in get_libraries(chip, include_asic=False):
            lib_key = ('library', item, *key)
            if __is_key_valid(chip, *lib_key):
                keys.append(lib_key)

    for key in keys:
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(key),
                 step=step, index=index)


def get_input_files(chip, *key, add_library_files=True):
    '''
    Returns a list of files from the key input and includes files
    from libraries if requested.

    Args:
        chip (Chip): Chip object
        key (list): Key to collect files from
        add_library_files (bool): When True, files from library keys
            will be included
    '''

    files = []
    for key in __get_keys(chip, *key, include_library_files=False):
        step, index = __get_step_index(chip, *key)
        files.extend(chip.find_files(*key, step=step, index=index))

    if add_library_files:
        for item in get_libraries(chip, include_asic=False):
            lib_key = ('library', item, *key)
            if __is_key_valid(chip, *lib_key):
                step, index = __get_step_index(chip, *lib_key)
                files.extend(chip.find_files(*lib_key, step=step, index=index))

    return __remove_duplicates(chip, files, list(key))


def has_input_files(chip, *key, check_library_files=True):
    '''
    Returns true if the specified key is set.

    Args:
        chip (Chip): Chip object
        key (list): Key to check
        check_library_files (bool): When True, files from library keys
            will be checked
    '''
    if __get_keys(chip, *key, include_library_files=check_library_files):
        return True

    return False


def __remove_duplicates(chip, values, type):
    new_values = []
    for v in values:
        if v not in new_values:
            new_values.append(v)
        else:
            chip.logger.warning(f"Removing duplicate {type}: {v}")
    return new_values


def __get_step_index(chip, *key):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    if chip.get(*key, field='pernode').is_never():
        step = None
        index = None

    return step, index


def __is_key_valid(chip, *key):
    if chip.valid(*key):
        step, index = __get_step_index(chip, *key)
        if chip.get(*key, step=step, index=index):
            return True
    return False


def __get_keys(chip, *key, include_library_files=True):
    keys = []
    if __is_key_valid(chip, *key):
        keys.append(key)

    if include_library_files:
        for item in get_libraries(chip):
            lib_key = ['library', item, *key]
            if __is_key_valid(chip, *lib_key):
                keys.append(tuple(lib_key))

    return keys


def __assert_support(chip, opt_keys, supports):
    from siliconcompiler import SiliconCompilerError

    if not supports:
        supports = []

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)
    for opt, vals in opt_keys.items():
        val_list = ', '.join([str(list(v)) for v in vals])
        if opt not in supports and val_list:
            msg = f'{tool}/{task} does not support [\'option\', \'{opt}\']'
            if len(vals) != 1 or len(vals[0]) != 2:
                msg += f', the following values will be ignored: {val_list}'
                pass
            chip.logger.warning(msg)

    for opt in supports:
        if opt not in opt_keys:
            raise SiliconCompilerError(
                f'{tool}/{task} is requesting support for {opt}, which does not exist',
                chip=chip)


def __get_frontend_option_keys(chip):
    opts = {
        'ydir': __get_keys(chip, 'option', 'ydir'),
        'vlib': __get_keys(chip, 'option', 'vlib'),
        'idir': __get_keys(chip, 'option', 'idir'),
        'define': __get_keys(chip, 'option', 'define'),
        'libext': __get_keys(chip, 'option', 'libext'),
        'param': []  # Only from 'option', no libraries
    }

    for param in chip.getkeys('option', 'param'):
        opts['param'].append(('option', 'param', param))

    return opts


def add_frontend_requires(chip, supports=None):
    '''
    Adds keys to the require list for the frontend task and checks if
    options are set, which the current frontend does not support.

    Args:
        chip (Chip): Chip object
        supports (list): List of ['option', '*'] which the frontend supports
    '''
    opt_keys = __get_frontend_option_keys(chip)
    __assert_support(chip, opt_keys, supports)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    for libkey in __get_library_keys(chip):
        chip.add('tool', tool, 'task', task, 'require', ','.join(libkey), step=step, index=index)

    for opt in supports:
        for key in opt_keys[opt]:
            chip.add('tool', tool, 'task', task, 'require', ','.join(key), step=step, index=index)


def get_frontend_options(chip, supports=None):
    '''
    Returns a dictionary of options set for the frontend and checks if
    options are set, which the current frontend does not support.

    Args:
        chip (Chip): Chip object
        supports (list): List of ['option', '*'] which the frontend supports
    '''
    opt_keys = __get_frontend_option_keys(chip)
    __assert_support(chip, opt_keys, supports)

    params = opt_keys['param']
    del opt_keys['param']

    opts = {}
    for opt, keys in opt_keys.items():
        opts[opt] = []
        for key in keys:
            sc_type = chip.get(*key, field='type')
            step, index = __get_step_index(chip, *key)
            if 'file' in sc_type or 'dir' in sc_type:
                opts[opt].extend(chip.find_files(*key, step=step, index=index))
            else:
                opts[opt].extend(chip.get(*key, step=step, index=index))

        opts[opt] = __remove_duplicates(chip, opts[opt], ['option', opt])

    opts['param'] = []
    for key in params:
        param = key[-1]
        value = chip.get(*key)
        opts['param'].append((param, value))

    return opts


def find_incoming_ext(chip, support_exts, default_ext):
    from siliconcompiler.utils import get_file_ext

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    flow = chip.get('option', 'flow')

    for input_step, input_index in chip.get('flowgraph', flow, step, index, 'input'):
        tool, task = get_tool_task(chip, input_step, input_index, flow=flow)
        output_exts = {get_file_ext(f): f for f in chip.get('tool', tool, 'task', task, 'output',
                                                            step=input_step, index=input_index)}
        # Search the supported order
        for ext in support_exts:
            if ext in output_exts:
                if output_exts[ext].lower().endswith('.gz'):
                    return f'{ext}.gz'
                return ext

    for ext in support_exts:
        for fileset in chip.getkeys('input'):
            if chip.valid('input', fileset, ext):
                return ext

    # Nothing found return the default
    return default_ext


def pick_key(chip, check_keys, step=None, index=None):
    if not step:
        step = chip.get('arg', 'step')
    if not index:
        index = chip.get('arg', 'index')

    for key in check_keys:
        if chip.valid(*key):
            check_step = step
            check_index = index

            if chip.get(*key, field='pernode').is_never():
                check_step = None
                check_index = None

            check_value = chip.get(*key, step=check_step, index=check_index)
            if check_value:
                return key, check_value

    return None, None


def input_provides(chip, step, index, flow=None):
    from siliconcompiler import NodeStatus

    if not flow:
        flow = chip.get('option', 'flow')

    pruned_nodes = chip.get('option', 'prune')

    nodes = chip.get('flowgraph', flow, step, index, 'input')
    inputs = {}
    for in_step, in_index in nodes:
        if (in_step, in_index) in pruned_nodes:
            # node has been pruned so will not provide anything
            continue

        if chip.get('record', 'status', step=in_step, index=in_index) == \
                NodeStatus.SKIPPED:
            for file, nodes in input_provides(chip, in_step, in_index, flow=flow).items():
                inputs.setdefault(file, []).extend(nodes)
            continue
        tool, task = get_tool_task(chip, in_step, in_index, flow=flow)

        for output in chip.get('tool', tool, 'task', task, 'output',
                               step=in_step, index=in_index):
            inputs.setdefault(output, []).append((in_step, in_index))

    return inputs


def input_file_node_name(filename, step, index):
    from siliconcompiler.utils import get_file_ext

    file_type = get_file_ext(filename)

    base = filename
    total_ext = []
    ext = None
    while ext != file_type:
        base, ext = os.path.splitext(base)
        ext = ext[1:].lower()
        total_ext.append(ext)

    total_ext.reverse()

    return f'{base}.{step}{index}.{".".join(total_ext)}'


def add_common_file(chip, key, file):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'file', key,
             f'tools/_common/{file}',
             step=step, index=index,
             package='siliconcompiler')
    chip.add('tool', tool, 'task', task, 'require',
             ','.join(['tool', tool, 'task', task, 'file', key]),
             step=step, index=index)


###########################################################################
def get_tool_task(chip, step, index, flow=None):
    '''
    Helper function to get the name of the tool and task associated with a given step/index.
    '''
    if not flow:
        flow = chip.get('option', 'flow')

    tool = chip.get('flowgraph', flow, step, index, 'tool')
    task = chip.get('flowgraph', flow, step, index, 'task')
    return tool, task


def get_tool_tasks(chip, tool):
    tool_dir = os.path.dirname(tool.__file__)
    tool_name = tool.__name__.split('.')[-1]
    tool_base_module = tool.__name__.split('.')
    if not tool.__file__.endswith('__init__.py'):
        tool_base_module = tool_base_module[0:-1]

    task_candidates = []
    for task_mod in pkgutil.iter_modules([tool_dir]):
        if task_mod.name.startswith('_'):
            continue

        if task_mod.name == tool_name:
            continue
        task_candidates.append(task_mod.name)

    tasks = []
    for task in sorted(task_candidates):
        task_module = '.'.join([*tool_base_module, task])
        if getattr(chip._load_module(task_module), 'setup', None):
            tasks.append(task)

    return tasks


#######################################
def record_metric(chip, step, index, metric, value, source, source_unit=None):
    '''
    Records a metric from a given step and index.

    This function ensures the metrics are recorded in the correct units
    as specified in the schema, additionally, this will record the source
    of the value if provided.

    Args:
        step (str): step to record the metric into
        index (str): index to record the metric into
        metric (str): metric to record
        value (float/int): value of the metric that is being recorded
        source (str): file the value came from
        source_unit (str): unit of the value, if not provided it is assumed to have no units

    Examples:
        >>> record_metric(chip, 'floorplan', '0', 'cellarea', 500.0, 'reports/metrics.json', \\
            source_units='um^2')
        Records the metric cell area under 'floorplan0' and notes the source as
        'reports/metrics.json'
    '''
    from siliconcompiler import units

    metric_unit = None
    if chip.schema.has_field('metric', metric, 'unit'):
        metric_unit = chip.get('metric', metric, field='unit')

    if metric_unit:
        value = units.convert(value, from_unit=source_unit, to_unit=metric_unit)

    chip.set('metric', metric, value, step=step, index=index)

    if source:
        flow = chip.get('option', 'flow')
        tool, task = get_tool_task(chip, step, index, flow=flow)

        chip.add('tool', tool, 'task', task, 'report', metric, source, step=step, index=index)


def has_pre_post_script(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    return chip.get('tool', tool, 'task', task, 'prescript', step=step, index=index) or \
        chip.get('tool', tool, 'task', task, 'postscript', step=step, index=index)



from siliconcompiler import utils
from siliconcompiler.tools import slang
from siliconcompiler.tools._common import \
    add_require_input, add_frontend_requires, get_tool_task, has_input_files


def setup(chip):
    '''
    Elaborate verilog design files and generate a unified file.
    '''
    slang.setup(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'threads', utils.get_cores(chip),
             clobber=False, step=step, index=index)

    add_require_input(chip, 'input', 'rtl', 'verilog')
    add_require_input(chip, 'input', 'rtl', 'systemverilog')
    add_frontend_requires(chip, ['ydir', 'idir', 'vlib', 'libext', 'define', 'param'])

    chip.set('tool', tool, 'task', task, 'stdout', 'destination', 'output', step=step, index=index)
    chip.set('tool', tool, 'task', task, 'stdout', 'suffix', 'v', step=step, index=index)

    chip.set('tool', tool, 'task', task, 'output', __outputfile(chip), step=step, index=index)


def runtime_options(chip):
    options = slang.common_runtime_options(chip)
    options.extend([
        "--preprocess",
        "--comments",
        "--ignore-unknown-modules",
        "--allow-use-before-declare"
    ])

    return options


def __outputfile(chip):
    is_systemverilog = has_input_files(chip, 'input', 'rtl', 'systemverilog')
    if is_systemverilog:
        return f'{chip.top()}.sv'
    return f'{chip.top()}.v'



'''
Graphviz is open source graph visualization software.
Graph visualization is a way of representing structural information as diagrams
of abstract graphs and networks.
It has important applications in networking, bioinformatics, software engineering,
database and web design, machine learning, and in visual interfaces for other
technical domains.
'''


def make_docs(chip):
    return chip



'''
Magic is a chip layout viewer, editor, and circuit verifier with
built in DRC and LVS engines.

Documentation: http://opencircuitdesign.com/magic/userguide.html

Installation: https://github.com/RTimothyEdwards/magic

Sources: https://github.com/RTimothyEdwards/magic
'''

import gzip
import shutil
import os
from siliconcompiler.tools._common import input_provides, get_tool_task
from siliconcompiler import utils


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    from siliconcompiler.targets import freepdk45_demo
    chip.use(freepdk45_demo)


################################
# Setup Tool (pre executable)
################################
def setup(chip):
    ''' Setup function for 'magic' tool
    '''

    tool = 'magic'
    refdir = 'tools/' + tool
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    # magic used for drc and lvs
    # if step not in ('drc', 'extspice'):
    #    raise ValueError(f"Magic tool doesn't support step {step}.")
    script = 'sc_magic.tcl'

    chip.set('tool', tool, 'exe', tool)
    chip.set('tool', tool, 'vswitch', '--version')
    chip.set('tool', tool, 'version', '>=8.3.196', clobber=False)
    chip.set('tool', tool, 'format', 'tcl')

    chip.set('tool', tool, 'task', task, 'threads', utils.get_cores(chip),
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'refdir', refdir,
             step=step, index=index,
             package='siliconcompiler', clobber=False)
    chip.set('tool', tool, 'task', task, 'script', script,
             step=step, index=index, clobber=False)

    # set options
    options = []
    options.append('-noc')
    options.append('-dnull')
    chip.set('tool', tool, 'task', task, 'option', options, step=step, index=index, clobber=False)

    design = chip.top()
    if f'{design}.gds' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', f'{design}.gds', step=step, index=index)
    else:
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['input', 'layout', 'gds']),
                 step=step, index=index)

    chip.set('tool', tool, 'task', task, 'regex', 'errors', r'^Error',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'regex', 'warnings', r'warning',
             step=step, index=index, clobber=False)


################################
# Version Check
################################
def parse_version(stdout):
    return stdout.strip('\n')


def process_file(file_type, chip, *key):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    if chip.get(*key, field='pernode').is_never():
        files = chip.find_files(*key)
    else:
        files = chip.find_files(*key, step=step, index=index)

    for file in files:
        if file.lower().endswith('.gz'):
            new_file_name = f'inputs/sc_{utils.get_hashed_filename(file[:-3])}'

            with gzip.open(file, 'rt', encoding="utf-8") as fin:
                with open(new_file_name, 'w') as fout:
                    fout.write(fin.read().encode("ascii", "ignore").decode("ascii"))
        else:
            new_file_name = f'inputs/sc_{utils.get_hashed_filename(file)}'
            shutil.copy(file, new_file_name)

        chip.add('tool', tool, 'task', task, 'file', f'read_{file_type}',
                 os.path.join(chip.getworkdir(step=step, index=index), new_file_name),
                 step=step, index=index)


##################################################
if __name__ == "__main__":

    chip = make_docs()
    chip.write_manifest("magic.json")



from siliconcompiler.tools.builtin import _common
from siliconcompiler.tools.builtin import nop
from siliconcompiler.tools.builtin.builtin import set_io_files


def setup(chip):
    '''
    Merges outputs from a list of input tasks.
    '''

    set_io_files(chip)


def _select_inputs(chip, step, index):
    return _common._select_inputs(chip, step, index)


def _gather_outputs(chip, step, index):
    return nop._gather_outputs(chip, step, index)


def run(chip):
    return _common.run(chip)


def post_process(chip):
    _common.post_process(chip)
